#! /tvbin/tivosh
# Copyright (c) 2001 TiVo Inc.
proc finishSoftwareInstallation {} {
    
    global env

    set db [dbopen]

    transaction {
        # For backwards compatibility with old databases, default
        # to "master" of there is no explicit setting.
        set name master

        # Don't use the Setup itcl object here.  Keep this script
        # lean and mean.
        # The ServiceInfo and SwSystemName fields are optional, and may not
        # be present.  This will only change name from "master" to something
        # else if the field is present in the db
        catch {
            set serviceConfig   [db $db open /State/ServiceConfig]
            set name            [dbobj $serviceConfig  get SwSystemName]
            if { $name == "" } {
                set name master
            }
        }

        set sw      [db $db open /SwSystem/$name]
        set active  [dbobj $sw get Active]

	# If an emergency reinstall was requested, pretend that the
	# current system software is not active

	catch {
	    if { $env(EMERGENCY_REINSTALL) == 1 } {
		puts "Emergency software reinstallation requested"
		set active 0
	    }
	}
    }

    if { $active != 1 } {
        puts "Installing software from SwSystem $name"
	catch {exec /tvbin/osdwriter /tvbin/InstallingSoftware.png }
        exec $env(TIVO_ROOT)/tvbin/tivosh $env(TIVO_ROOT)/tvbin/installSw.itcl $name >&@ stdout
        ## should not come back.
        error "System should have rebooted"
    } else {
        puts "SwSystem $name is already active, nothing to do."
    }
}

#
# If there is a new SwSystem object there, install it.
#

puts "Checking new software installation"
if [catch {finishSoftwareInstallation} res ] {
    puts "Could not check software installation: $res"
}


