#!/tvbin/tivosh
##############################################################################
# mfs_ftp - Copyright (C) 2001-2003  Riley Cassel<RC3105@hotmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# For a copy of the GNU General Public License write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##############################################################################
proc init_mfs_ftp { } {
global  argv info control_socket
 set info(ltime) [clock seconds]
 outc "entering init_mfs_ftp"
 set info(version) "1.2.5.9" ;# 0-10, 10 == slowest
 # incorporates zobetron's timezone auto-detect proc - still clunky but
 # functional - still working on tyx details but works with tyedit beta 3
 # includes a dump_basic_xml for extracting tmf's from a lightly toasted mfs
 set info(ithrottle)	0	;# inserts
 set info(ethrottle)	0	;# extracts - not implemented yet
 set info(multi_c_threads) 1
 set info(pingpong) 300000
# set info(tzos) [expr 3600 * 6] ;# central standard timezone offset
set info(tzos) [get_tzoffset_one]
 catch { if { [file exist /var/mfs_ftp/abort.txt] } { exec /bin/mv abort.txt abort.doc } }
 set info(port) [lindex $argv 0]
 if { $info(port) == "" } { set info(port) 3105 }
 set info(wrapped) [lindex $argv 1]
 set info(localip) "127,0,0,1"
 set info(type) "I"
 background
 ping
 versioncheck
 set info(cached_rec_fsids) ""
 set info(rec_fsids) "0"
 if { $info(ithrottle) > 10 } { set info(ithrottle) 10 }
 if { $info(ithrottle) < -1 } { set info(ithrottle) -1 }
 set info(ithrottlems) [expr $info(ithrottle) * 10]
 outc "insert throttle is\t$info(ithrottle), == $info(ithrottlems) ms "
 if { $info(ethrottle) > 10 } { set info(ethrottle) 10 }
 if { $info(ethrottle) < -1 } { set info(ethrottle) -1 }
 set info(ethrottlems) [expr $info(ethrottle) * 10]
 outc "extract throttle is\t$info(ethrottle), == $info(ethrottlems) ms "
 reset_mfs_ftp
 build_dir_LIST
 update_rec_fsids
 event register $TmkEvent::EVT_DATA_CHANGED data_changed_event
}


proc reset_mfs_ftp { } {
global list argv info control_socket
 outc "entering reset_mfs_ftp"
 catch {close $info(lastsock)}
 catch {close $control_socket(main)}
 set info(dir) "/"
 set info(fmt) "/"
 set info(mode) normal
 setup_control_socket $info(port)
}


proc background { } {
global info
 outc "entering background"
 if { $info(wrapped) != "background" } {
#	exec /var/mfs_ftp/mfs_ftp.tcl $info(port) background >&@ stdout &
	exec /var/mfs_ftp/mfs_ftp.tcl $info(port) background >/var/mfs_ftp/ftp.log &
	exit
 }
}


proc	ABOR		{ args fsock } {
outc "entering ABOR - abort"
 set option1 [lindex $args 1] 
 tee $fsock "200 Transfer Aborted"
 tee $fsock "200 Transfer Aborted"
}


proc	CDUP	{ args fsock } {
outc "entering CDUP"
global info
 set info(dir) "/"
 tee $fsock "250 Directory change successful."
}


proc 	DELE		{ args fsock line } {
outc "entering DELE - delete a recording"
 set line [string range $line 5 end]
 set choplength [string last / $line]
 set line [string range $line [expr $choplength +1] end]
 outc "line is now \"$line\""
 set args [split $line \ ]
 set recfsid [lindex $args 0]
 set recfsid [dropslash $recfsid] 
 if {![isInt $recfsid]} {
	tee $fsock "550 error, can't find \"$recfsid\""
	return
 }
 outc "attempting to delete fsid \"$recfsid\""
 global db
 set db [dbopen]
 set canceldate [expr [clock seconds] / 86400]
 set canceltime [expr [clock seconds] % 86400]
 set delay 0
 RetryTransaction {
	after $delay 
	set delay 100
	set rec [db $db openid $recfsid]
	set state [dbobj $rec get State]
	if { $state == 3 } {
        	tee $fsock "550 error, can't delete a recording in progress"
		return
	} else {
		dbobj $rec set CancelReason 12
		dbobj $rec set DeletionDate $canceldate
		dbobj $rec set DeletionTime $canceltime
		set errorstring [dbobj $rec get ErrorString]
		set elength [string length $errorstring]
		if { $elength > 0 } {
			set errorstring [string trim $errorstring "\{\}"]
			dbobj $rec set ErrorString "$errorstring Deleted by user"
		} else {
			dbobj $rec set ErrorString "Deleted by user"
		}
		dbobj $rec set State 5
	}
 }
 dbclose $db
 tee $fsock "250 DELE command successful."
 update_rec_fsids
}


proc	CWD	{ args fsock } {
global info
 set nwd [lindex $args 1]
 set nwd [dropslash $nwd]
    switch $nwd {
	pheonix { after 3000 pheonix_mfs_ftp stdout ChangeWorkingDirectory }
	fork 	{ fork_mfs_ftp }
	txt	{ set info(dir) "txt" ; set info(fmt) "txt" }
	tmf	{ set info(dir) "tmf" ; set info(fmt) "tmf" }
	shutdown { after 3000 exitproc stdout ChangeWorkingDirectory }
	tyx	{ set info(dir) "tyx" ; set info(fmt) "tyx" }
	xml	{ set info(dir) "xml" ; set info(fmt) "xml" }
	ty	{ set info(dir) "ty" ; set info(fmt) "ty" }
	/	{ set info(dir) "/" ; set info(fmt) "/" }
	default { set info(dir) "/" ; set info(fmt) "/" }
    }
 tee $fsock "250 Directory change successful."
}


proc	LIST		{ line fsock } {
outc "entering LIST "
global info
 tee $fsock "150 Opening data connection for file list."
 fconfigure $info(dc) -translation binary
 build_list $line
 puts -nonewline $info(dc) $info(list)
 tee $fsock "226 Transfer complete."
 close $info(dc)
}


proc build_list { line } {
outc "entering build_list "
global info
 if { $info(dir) == "/" } {
	set info(list) $info(dir_list)
 } else {
	set info(list) ""
	set atbs "-r--r--r-- 1 0 0"
	set temp ""
	foreach fsid $info(rec_fsids) {
		set tsize $info($fsid,tsize)
		if { $info(fmt) == "xml" } { set tsize 8192 }
		if { $info(fmt) == "txt" } { set tsize 512 }
		set temp "$atbs $tsize $info($fsid,rec_info).$info(fmt)\r\n"
		append temp $info(list)
		set info(list) $temp
		}
 }
}


proc	NLST		{ line fsock } {
outc "entering NLST "
global info
 tee $fsock "150 Opening data connection for file list."
 fconfigure $info(dc) -translation binary
 build_short_list $line
 puts -nonewline $info(dc) $info(list)
 tee $fsock "226 Transfer complete."
 close $info(dc)
}
          

proc build_short_list { line } {
outc "entering build_short_list "
global info
 set info(list) ""
 set temp ""
 foreach fsid $info(rec_fsids) {
	set tsize $info($fsid,tsize)
	if { $info(fmt) == "xml" } { set tsize 8192 }
	if { $info(fmt) == "txt" } { set tsize 512 }
	set temp "$info($fsid,short_info).$info(fmt)\r\n"
	if [ string length [lindex $line 1] ] {
    		if [ string match [lindex $line 1] $temp ] {
        		append temp $info(list); set info(list) $temp
    		}
	} else {
		append temp $info(list); set info(list) $temp
	}
 }
}

                   
proc data_changed_event { type subtype } {
#outc "entering data_changed_event - \"$type $subtype\" data change event"
global info
  switch $subtype {
	0	{ update_rec_fsids }
	1	{ update_rec_fsids }
	2	{ update_rec_fsids }
	4	{ update_rec_fsids }
	default	{ }
  }
}


proc update_rec_fsids { } {
outc "entering update_rec_fsids "
global info db
set info(rec_fsids) ""
 set db [dbopen]
 outc "getting current rec_fsids list "
 ForeachMfsFile fsid name type "$info(mfspath)" "" {
	lappend info(rec_fsids) $fsid
 }
 dbclose $db
 if { $info(rec_fsids) != $info(cached_rec_fsids) } {
	outc "rec_fsids doesn't match cached_rec_fsids, deleting cached info"
	foreach fsid $info(cached_rec_fsids) {
		unset info($fsid,rec_info)
		unset info($fsid,tsize)
		unset info($fsid,short_info)
		unset info($fsid,parts)
		unset info($fsid,state)
	}
	set info(cached_rec_fsids) $info(rec_fsids)
	build_rec_LIST
 } else {
	outc "new/old lists match - no update necesary"
 }
}


proc build_dir_LIST { } {
outc "entering build_dir_LIST"
global info
 set info(dir_list) ""
 set date [clock format [expr [clock seconds] +$info(tzos)] -format "%b %d %Y" ]
 append info(dir_list) "drw-rw-r-- 2 0 0 1024      $date tyx\r\n"
 append info(dir_list) "drw-rw-r-- 2 0 0 1024      $date tmf\r\n"
 append info(dir_list) "dr--r--r-- 2 0 0 1024      $date ty\r\n"
 append info(dir_list) "dr--r--r-- 2 0 0 1024      $date xml\r\n"
 append info(dir_list) "dr--r--r-- 2 0 0 1024      $date txt\r\n"
 if { $info(wrapped) != "wrapped" } {
	append info(dir_list) "drwxrwxr-x 2 0 0 1024      $date pheonix\r\n"
 }
 if { $info(multi_c_threads) == 0 } {
	append info(dir_list) "drwxrwxr-x 2 0 0 1024      $date fork\r\n"
 }
 append info(dir_list) "drwxrwxr-x 2 0 0 1024      $date shutdown\r\n"
}


proc	build_rec_LIST	{ } {
outc "entering build_rec_list "
global info
 set date "Dec 25 2001"
 outc "from mfs \"$info(mfspath)\" "
 set atbs "-r--r--r-- 1 0 0"
 set db [dbopen]
 foreach fsid $info(rec_fsids) {
  set tsize 0 ; set file_size 0 ; set ip "" ; set delay 0
    RetryTransaction {
	after $delay
	set delay 100
	set rec [db $db openid $fsid]
	set state [dbobj $rec get State]
	set info($fsid,state) State
	if {$state != 4} { set ip "Recording - " }
	set showing [dbobj $rec get Showing]
	set recsec [expr [dbobj $showing get Date] * 86400 + [dbobj $showing get Time] + $info(tzos)]
	set date [clock format $recsec -format "%b %1d %H:%2M" ]
#	set date [clock format $recsec -format "%b %1d %I:%2M" ]
	set program [dbobj $showing get Program]
	set title [strim [dbobj $program get Title]]
	set episode [strim [dbobj $program get EpisodeTitle]]
	set tsize [dbobj $rec get StreamFileSize]
	set info($fsid,parts) ""
	set parts [dbobj $rec get Part]
	foreach part $parts {
		set file [dbobj $part get File]
		lappend info($fsid,parts) $file
	}
	if { $tsize != "" } {
		set tsize [expr $tsize * 1024]
	} else {
		foreach part $parts {
			set file [dbobj $part get File]
			if { [catch { mfs streamsize $file } part_size ] } {
				set tsize 536870912
			} else {
				set blocksize [lindex $part_size 0]
				set blocks [lindex $part_size 1]
				set file_size [expr $blocks * $blocksize]
				if {$file_size == 0} { set file_size 524288 }
				set tsize [expr $tsize + $file_size]
			};#end if catch else
		};#end foreach
	};#end if tzise else
    };#end retrytransaction
  set tsize [format "%#010u" $tsize]
  set info($fsid,tsize) $tsize
  set info($fsid,rec_info) "$date $fsid $ip$title - $episode"
  set info($fsid,short_info) "$fsid $ip$title - $episode"
  puts -nonewline stdout "." ; flush stdout
# outc "$info($fsid,short_info) $info($fsid,parts) "
 };#end of foreach fsid
 dbclose $db
 puts stdout ""
 outc "build_rec_LIST complete "
}
          

proc	NOOP		{ args fsock } {
set option1 [lindex $args 1] 
tee $fsock "200 Command okay."
}


proc	PASS	{ args fsock } {
set option1 [lindex $args 1] 
tee $fsock "230 Running in TiVo Mode."
}


proc	PASV		{ args fsock } {
outc "entering pasv"
global pdcs info
 outc "closing info(dc) (just in case)"
 catch { close $info(dc) }
 set info(mode) "passive"
 set option1 [lindex $args 1] 
 set ipaddr $info(localip)
 set lport [expr $info(port) - 1]
 set lp1 [expr $lport / 256]
 set lp2 [expr $lport - ($lp1 * 256)]
 set lportftp "$lp1,$lp2"
 set pdcs(main) [socket -server newpdc $lport]
 outc "passive dc server socket listening on $info(localip) $lport"
 tee $fsock "227 Entering Passive Mode ($ipaddr,$lportftp)."
};#end of proc


proc newpdc {sock addr port} {
global pdcs info
 #outc "newpdc server socket accept $sock from address $addr port $port"
 #outc "closing server data socket to new connections"
	close $pdcs(main)
 #outc "setting info(dc) to passive data channel $sock"
	set info(dc) $sock
	fconfigure $sock -translation binary
}


proc	PORT		{ line fsock } {
outc "entering PORT"
global info pdcs
 set info(mode) normal
 set args [split $line \ ,]
 set ip1 [lindex $args 1]
 set ip2 [lindex $args 2]
 set ip3 [lindex $args 3] 
 set ip4 [lindex $args 4] 
 set ip5 [lindex $args 5] 
 set info(clientip) $ip1.$ip2.$ip3.$ip4
 set portoffset [lindex $args 6]
 outc "$ip1 $ip2 $ip3 $ip4 $ip5 $portoffset"
 set info(cport) [expr $ip5 * 256 + $portoffset]
 outc "opening connection to $info(clientip) : $info(cport)"
 set info(dc) [socket $info(clientip) $info(cport)]
 fconfigure $info(dc) -buffersize 131072 -translation binary
 tee $fsock "200 PORT command successful."
}


proc	PWD 		{ args fsock } {
global info
set option1 [lindex $args 1] 
	if { $info(dir) == "/" } {
		tee $fsock "257 \"$info(dir)\" is current directory."
	} else { 
		tee $fsock "257 \"/$info(dir)\" is current directory."
	}
}


proc	QUIT	{ fsock } {
global info
	tee $fsock "221 Server Closing Control Connection by client request"
	after 1 reset_mfs_ftp
}


#mbm's tcl object dumper 
proc dump_object { obj {id ""} {depth 10} {prefix ""}} {
  if { $depth == 0 } {return "** max depth **\n"}
  set output "";
  set fsid [dbobj $obj fsid]
  set subid [dbobj $obj subobjid]
  if { [info exists ::seen($fsid/$subid)] } {
    return {}
  }
  set ::seen($fsid/$subid) "$id"
  set otype [dbobj $obj type]
  if {[dbobj $obj primary]} {
    append output "$prefix<Object type=\"$otype\" id=\"$id\">\n"
    set endtag "</Object>"
  } else {
    append output "$prefix<SubObject type=\"$otype\" id=\"$id\">\n"
    set endtag "</SubObject>"
  }
  foreach attr [lsort [dbobj $obj attrs]] {

outc "attr \"$attr\""

    set attrs [dbobj $obj get $attr]
    set atype [dbobj $obj attrtype $attr]
    switch -glob $attr {
      0xe0039 -
      ActualShowing -
      BitRate -
      Bookmark -
      CancelReason -
      Deletion* -
      Expiration* -
      ErrorString -
      File -
      ProgramSource -
      Index* -
      NSecondsWatched -
      NVisit -
      RecordingBehavior -
      Score -
      SelectionType -
      State -
      Version {continue}
    }
    foreach a $attrs {
      switch $atype {
        object { append output [dump_object $a $attr [expr $depth-1] "$prefix "]; set a "" }
      }
      if { $a != "" } {
        append output "$prefix  <$attr>$a</$attr>\n"
      }
    }                                             
  }
  append output "$prefix$endtag\n"
  return "$output"
}


proc dump_xml { obj } {
outc "entering dump_xml"
  global db
  set version [dbobj [db $db open "/SwSystem/ACTIVE"] get Name]
  set output "<?xml version=\"1.0\" tivoversion=\"$version\"?>\n"
  append output [dump_object $obj "_top"]
  unset ::seen
  return $output
}


proc dump_basic_xml { rec } {
outc "entering dump_basic_xml"
  global db
  set version [dbobj [db $db open "/SwSystem/ACTIVE"] get Name]
  set output "<?xml version=\"1.0\" tivoversion=\"$version\"?>\n"

        set showing [dbobj $rec get Showing]
        set program [dbobj $showing get Program]
	set station [dbobj $showing get Station]

	catch { set Title [strim [dbobj $program get Title]] }
	if { $Title == "" } { set Title "no Title available" }
	append output "<Title>$Title</Title>\n"

	catch { set duration [dbobj $showing get Duration] }
	if { $duration == "" } { set duration "1800" }
        append output "<Duration>$duration</Duration>\n" 

	catch { set date [dbobj $showing get Date] }
	if { $date == "" } { set date "0" }
        append output "<Date>$date</Date>\n"

	catch { set Description [dbobj $program get Description] }
	if { $Description == "" } { set Description "no Description available" }
	append output "<Description>$Description</Description>\n"

	catch { set EpisodeTitle [strim [dbobj $program get EpisodeTitle]] }
	if { $EpisodeTitle == "" } { set EpisodeTitle "no EpisodeTitle available" }
	append output "<EpisodeTitle>$EpisodeTitle</EpisodeTitle>\n"

	catch { set CallSign [dbobj $station get CallSign] }
	if { $CallSign == "" } { set CallSign "bogus" }
	append output "<CallSign>$CallSign</CallSign>\n"

	catch { set Name [dbobj $station get Name] }
	if { $Name == "" } { set Name "NA" }
	append output "<Name>$Name</Name>\n"

	catch { set Time [dbobj $showing get Time] }
	if { $Time == "" } { set Time "0" }
	append output "<Time>$Time</Time>\n"

	catch { set StartDate [dbobj $rec get StartDate] }
	if { $StartDate == "" } { set StartDate "0" }
	append output "<StartDate>$StartDate</StartDate>\n"

catch { set StopDate [dbobj $rec get StopDate] }
if { $StopDate == "" } { set StopDate "0" }
append output "<StopDate>$StopDate</StopDate>\n"

catch { set StartTime [dbobj $rec get StartTime] }
if { $StartTime == "" } { set StartTime "0" }
append output "<StartTime>$StartTime</StartTime>\n"

catch { set StopTime [dbobj $rec get StopTime] }
if { $StopTime == "" } { set StopTime [expr $StartTime + $Duration] }
append output "<StopTime>$StopTime</StopTime>\n"  

  
  return $output
}        




proc    RETR         { args fsock line } {
outc "entering RETR - retrieve video"
global info db recfsid
 set recfsid [dropslash [lindex $args 1 ]]
 if {![isInt $recfsid]} {
	tee $fsock "550 error, can't find \"$recfsid\""
	return
 }
 tee $fsock "150 About to open data connection."

 set info($recfsid,fsids) ""
 set delay 0
 set db [dbopen]
 RetryTransaction {
	after $delay
	set delay 100 
	set rec [db $db openid $recfsid]                      
	set parts [dbobj $rec get Part]
	foreach part $parts {
		set file [dbobj $part get File]
		lappend info($recfsid,fsids) $file
	}

	set info(xml) "[dump_xml $rec]"

# this is to extract tmf's where only minimal info is avail or there are
# wacky extra invalid attributes

#	set info(xml) "[dump_basic_xml $rec]"
#	set info(xml) "unknown xml info"

 }
 dbclose $db
 outc "fconfigure"
 fconfigure $info(dc) -translation binary

 switch $info(dir) {

   tyx	{
	outc "heading for buildtyx"
	buildtyx
	}

   txt	{
	parsexml $info(xml)
	build_txt_from_xml
	puts -nonewline $info(dc) $info(txt)
	}

   xml	{
	puts -nonewline $info(dc) $info(xml)
	}

   ty  {
#	outc "heading for buildtyx"
	buildty
	}

   tmf {
	set pout [open "|/var/mfs_ftp/mfs_tarstream -x -s $info($recfsid,fsids) >@$info(dc) 2>/dev/null " w]
	puts -nonewline $pout $info(xml)
	close $pout
	}

 };#end of switch

set xml ""
outc "\n"
close $info(dc)
tee $fsock "226 Transfer complete."
}


proc buildty { } {
global info recfsid
outc "entering buildty"
	set info(ty+) "################################"
	append info(ty+) $info(ty+); append info(ty+) $info(ty+);
	append info(ty+) $info(ty+); append info(ty+) $info(ty+);
	outc "fsid list is \"$info($recfsid,fsids)\""
	foreach fsid $info($recfsid,fsids) {
		outc "extracting part \"$fsid\""
		exec /var/mfs_ftp/mfs_stream -s $fsid >@$info(dc) 2>/dev/null
	}
	puts -nonewline $info(dc) $info(ty+)\n$info(xml)\n$info(ty+)
}       


proc buildtyx { } {
outc "entering buildtyx"
global info recfsid
 set numofparts 0
 set bytes_written 0
 foreach fsid $info($recfsid,fsids) {
	incr numofparts
 }
 outc "build showing.vro tarheader"
 set vro_th [build_th "\xf5\x46\x7A\xBD\x5fshowing.vro" "" [expr ($numofparts+1)*131072-512]]
 puts -nonewline $info(dc) $vro_th ; flush $info(dc)
 incr bytes_written 512
outc "build showing_xml_tp"
 set showing_xml_tp [build_th showing.xml $info(xml) "0"]
 append showing_xml_tp $info(xml)
 append showing_xml_tp [padeof [calc_tpp $info(xml) "0"]]
 set xml_size [string length $showing_xml_tp]
 puts -nonewline $info(dc) $showing_xml_tp
 flush $info(dc)
 incr bytes_written $xml_size
 set source_info_tp [build_th source_info.xml "" "512"]
 append source_info_tp [padspace 512]
 set sinfo_size [string length $source_info_tp]
 puts -nonewline $info(dc) $source_info_tp
 flush $info(dc)
 incr bytes_written $sinfo_size
 outc "figuring padding.null size"
 # 1536 = vro tarh + pnull th + chunks.tyh th
 set pnull_size [expr 131072 - 1536 - $xml_size - $sinfo_size]
 set pnull_th [build_th "vro_internal_padding.null" "" $pnull_size]
 set pnull_data [padspace $pnull_size]
 puts -nonewline $info(dc) $pnull_th
 puts -nonewline $info(dc) $pnull_data
 flush $info(dc)
 outc "generating chunks.tyh tarheader"
 set chunks_tyh_th [build_th chunks.tyh "" "[expr $numofparts * 131072]" ]
 puts -nonewline $info(dc) $chunks_tyh_th
 flush $info(dc)
 foreach fsid $info($recfsid,fsids) {
outc "writing header chunks"
	exec /var/mfs_ftp/mfs_export -c 131072 $fsid /var/mfs_ftp/fifo >&/dev/null &
	set tempchan [open /var/mfs_ftp/fifo r]
	fconfigure $tempchan -translation binary
#	set chunk [read $tempchan 131072] ;puts -nonewline $info(dc) $chunk
	fcopy $tempchan $info(dc) -size 131072 ; flush $info(dc)
	outc "flush fifo"
	while { [eof $tempchan] != 1 } {
		set junk [read $tempchan 1024]
	}
	close $tempchan
	after 1000 { catch { unset junk } }
 }
 set paddingsize [expr 130560 - $xml_size]
 set pnull_th [build_th "\xf5\x46\x7A\xBD\x5fpadding.null" "" $paddingsize]
 set size_diff [expr $paddingsize - [string length $pnull_data] ]
 set pnull_data_two [padspace $size_diff]
 puts -nonewline $info(dc) $pnull_th
 puts -nonewline $info(dc) $pnull_data
 puts -nonewline $info(dc) $pnull_data_two ; flush $info(dc)
 puts -nonewline $info(dc) $showing_xml_tp ; flush $info(dc)
 outc "building standard padding.null"
 set paddingsize 130048
 set pnull_th [build_th "\xf5\x46\x7A\xBD\x5fpadding.null" "" $paddingsize]
 set size_diff [expr $paddingsize - [string length $pnull_data] ]
 set pnull_data_two [padspace $size_diff]
 foreach fsid $info($recfsid,fsids) {
	outc "putting padding.null into info(dc)"
	puts -nonewline $info(dc) $pnull_th ; flush $info(dc)
	puts -nonewline $info(dc) $pnull_data ; flush $info(dc)
	puts -nonewline $info(dc) $pnull_data_two ; flush $info(dc)
	outc "copying fsid \"$fsid\" into info(dc)"
	set tempchan [open "|/var/mfs_ftp/mfs_tarstream -s $fsid 2>/dev/null " r]
	fconfigure $tempchan -translation binary
	outc "reading mfs_tarstream header"
	set tarh [read $tempchan 512]
	set bytes [expr [string range $tarh 124 134]]
	set chunks [expr int($bytes / 131072)]
	outc "$chunks chunks in \"$bytes\" bytes"
	outc "is actually [expr $chunks * 131072] bytes"
	puts -nonewline $info(dc) $tarh ; flush $info(dc)
	set chunknum 0
	set bytes_copied 0
	outc "while chunknum < chunks \n"
	while { $chunknum < $chunks } {
		if { [eof $tempchan] != 1 } {
			#set chunk [read $tempchan 131072]
			#puts -nonewline $info(dc) $chunk ; flush $info(dc)
			fcopy $tempchan $info(dc) -size 131072
			#set chunksize [string length $chunk]
			incr bytes_copied 131072
			puts -nonewline stdout "." ; flush stdout
		}
		incr chunknum
	}
	puts stdout ""
	outc "flushing tempchan"
	while { [eof $tempchan] != 1 } {
		set chunk [read $tempchan 1024]
	}
	close $tempchan
	outc "bytes actually copied was \"$bytes_copied\""
	outc ""
  }
  unset pnull_data ; unset pnull_data_two
}


proc build_th { name data bytes } {
#outc "entering build_th"
global info
 if { $bytes == "0" } { set bytes [string length $data] }
 set size [format "%#011o" $bytes]
 set tarh ""
 append tarh "$name"
 outc "building tarheader for \"$tarh\" "
 outc "\"$bytes\" bytes (\"$size\" in octal) / [expr $bytes + 512] total"
 set slen [string length $tarh]
 set namepad [expr 100 - $slen]
 append tarh [padeof $namepad] ; append tarh "0100644\x00"
 append tarh "0000000\x00" ; append tarh "0000000\x00"
 append tarh $size ; append tarh [padeof 1]
 append tarh "07632542044\x00" ; append tarh "        "
 append tarh "0" ; append tarh [padeof 100]
 append tarh "ustar  \x00" ; append tarh "tivo"
 append tarh [padeof 28] ; append tarh "tivo" ; append tarh [padeof 211]
 #outc "fixing to return tarh"
 set tarh [calc_checksum $tarh]
 return $tarh
}


proc calc_checksum { tarh } {
#outc "entering calc_checksum"
 set sum 0
 for { set i 0 } { $i < 512 } { incr i } {
	after 1
	set char [string range $tarh $i $i]
	binary scan $char H* charb
	set junk [scan $charb %x charval ]
	set sum [expr $sum + $charval]
 }
 set part_two "[format "%#07o" $sum]\x00"
#outc "sum is \"$sum\" octal \"$part_two\""
 set part_one "[string range $tarh 0 147 ]"
 set part_three [string range $tarh 156 511 ]
 set newtarh $part_one ; append newtarh $part_two ; append newtarh $part_three
return $newtarh
}


proc calc_tpp { data slen } {
outc "entering calc_tarpacket_padding"
if { $slen == "0" } { set slen [string length $data] }
 set blocks [expr $slen / 512]
 if { [expr $blocks * 512] < $slen } { incr blocks }
 set pbytes [expr $blocks * 512 - $slen]
 return $pbytes
}


proc padeof { num  } {
#outc "entering padeof - generating \"$num\" null chars"
set delay 100
if { $num > 512 } { set delay 250 }
set loops 0 ; set padsize 0 ; set ncsize 1 ; set padding ""
  while { [expr $num - $padsize] > 0 } {
	set nc \x00
	while { [expr ($num-$padsize) - $ncsize*2] > 0 } {
		append nc $nc
		set ncsize [string length $nc]
		incr loops
		after $delay
	}
	append padding $nc
	set padsize [string length $padding]
  }
#outc "padding size is \"$padsize\" in \"$loops\" loops"
return $padding
}      


proc padspace { num  } {
#outc "entering padspace - generating \"$num\" null chars"
set delay 100
if { $num > 512 } { set delay 250 }
set loops 0 ; set padsize 0 ; set ncsize 1 ; set padding ""
  while { [expr $num - $padsize] > 0 } {
	set nc \x20
	while { [expr ($num-$padsize) - $ncsize*2] > 0 } {
		append nc $nc
		set ncsize [string length $nc]
		incr loops
		after $delay
	}
	append padding $nc
	set padsize [string length $padding]
  }
#outc "padding size is \"$padsize\" in \"$loops\" loops"
return $padding
}


proc build_txt_from_xml { } {
outc "entering build_text_from_xml"
global info
 set info(txt) ""
 append info(txt) "$info(Title)\r\n"
 set secs [expr $info(Date)*3600*24 + $info(Time) + $info(tzos)]
 set temp [clock format $secs -format "(%Y)" ]
 append info(txt) "\"$info(EpisodeTitle)\" $temp \r\n"
 append info(txt) "$info(Description) \r\n\r\n"
 set secs [expr $info(Date) *3600*24 + $info(Time) +$info(tzos)]
 set temp [clock format $secs -format "%a %m/%d %I:%M %p"]
 append info(txt) "$temp $info(CallSign) \r\n"
 set temp [clock format $info(Duration) -format "%I:%M" ]
 append info(txt) "Duration: $temp \r\n"
 set secs [expr $info(OriginalAirDate) *3600*24 +$info(tzos)]
 set temp [clock format $secs -format "%m/%d/%Y" ]
 append info(txt) "Original Air Date: $temp \r\n"
}


proc outc { string } {
global info
 set elapsed "[expr [clock seconds] - $info(ltime)]"
 set elapsed [clock format $elapsed -format %H:%M:%S]
 puts stdout "$elapsed - $string"
}


proc tee { fsock string } {
 outc $string
 if { $fsock != "stdout" } { puts $fsock $string }
 if { $fsock != "stdin" } { puts stdout $string }
}


proc	STOR		{ args fsock line } {
# parse a tmf into fsid's & info & store it
outc "entering stor proc"
global info
 set args2 [split $line \ .]
	foreach item $args2 {
		set filext $item
	}

 if { $filext != "tmf" && $filext != "tyx" && $filext != "tar" } {
outc "wrong file extension"
tee $fsock "550 error, inserted files must be tmf tar or tyx (and just renaming 
a ty is a recipe for disaster)"
return
 }

 set option1 [lindex $args 1] 
 tee $fsock "150 Opening BINARY mode data connection for \"$option1\""
 set newrecording [ mkblankrec ]
 set xmldata [dumpxml]
 parsexml "$xmldata"
 setrecinfo $newrecording
 set info(finished) 0 ; set info(totalsize) 0
 set info(meg) 0 ; set info(10meg) 0
 while { [eof $info(dc)] != 1 } {
	tmf2fsid $newrecording
 }
 close $info(dc)
 outc "new recording $newrecording size is $info(totalsize)"
 tee $fsock "226 File transfer complete"
 outc "sending data changed event"
 event send $TmkEvent::EVT_DATA_CHANGED $TmkDataChanged::RECORDINGS_ON_DISK 0
 update_rec_fsids
}


proc tmf2fsid { fsid } {
global info
outc "entering tmf2fsid"
 set inf $info(dc)
 set tarh [read $info(dc) 512]
 if { $info(finished) == 1 } { return }
 set nlen [string first \x00 $tarh]
	if { $nlen == 0 } {
		 outc "end of file or invalid segment - leaving tmf2fsid"
		set info(finished) 1
		return
	} else {
		set nlen [expr $nlen - 1]
		set filename [string range $tarh 0 $nlen]
		outc "starting segment \"$filename\""
	}
 set bytes [expr [string range $tarh 124 134]]
 set blocks [expr int($bytes / 512)]
 set info(totalsize) [expr $info(totalsize) + $bytes/1024]
 if { [expr $blocks * 512] < $bytes } { incr blocks }
 set pbytes [expr $blocks * 512 - $bytes]
 set newfsid [AddPart $fsid $bytes]
 set outf [open "| /var/mfs_ftp/mfs_stdinsert $newfsid - >&/dev/null" w]
	fconfigure $outf -buffersize 131072 -translation binary

 if { $info(ithrottle) < 0 } {
	outc "fcopying $bytes bytes at a time - please wait"
fcopy $inf $outf -size $bytes
} else {
    set chunksize 131072
    outc "copying 1 meg per \".\""
    set stime  [expr [expr [clock seconds] % 86400] -1800]
    set chunknum 0
    for {set i 0} {$i < $bytes} {incr i $chunksize} {
	set s [min $chunksize [expr $bytes - $i]]
	fcopy $inf $outf -size $chunksize
#set chunk [read $inf $chunksize] ; puts -nonewline $outf $chunk
	if { [eof $info(dc)] == 1 } { set i $bytes }
# experimental change!
	incr chunknum
	if { $chunknum == 8 } {
		set chunknum 0
		incr info(meg)
		if { $info(meg) == 10 } {
			incr info(10meg) 10
puts -nonewline stdout "[format "%# 4u" $info(10meg)] \t" ; flush stdout
			set info(meg) 0
			after 100 
# exp delay added to improve tivo responsiveness during transfers
		} else {
			puts -nonewline stdout "." ; flush stdout
		}
	}
	after $info(ithrottlems)
    }
	set etime  [expr [expr [clock seconds] % 86400] -1800]
	set kilobytes [expr $bytes/1024]
	set totaltime [expr $etime - $stime]
	set datarate [expr $kilobytes / $totaltime]
	puts stdout " $datarate k / second"
}    
  set padding [read $inf $pbytes]
  close $outf
}


proc dumpxml {} {
outc "entering dumpxml"
global info inf pbytes tarh xmldata nlen filename
 set inf $info(dc)
 set tarh [read $inf 512]
 set nlen [expr [string first \x00 $tarh] -1]
 set filename [string range $tarh 0 $nlen]
 outc "segment is $filename"
 set bytes [expr [string range $tarh 124 134]]
 set blocks [expr int($bytes / 512)]
 if { [expr $blocks * 512] < $bytes } { incr blocks }
 set pbytes [expr $blocks * 512 - $bytes]
 set xmldata [read $inf $bytes]
 set padding [read $inf $pbytes]
 return $xmldata
}


proc parsexml { xmldata } {
outc "entering parsexml"
global info attributes
 set info(attributes) "Title Description EpisodeTitle StartDate StopDate 
StartTime StopTime Duration Date OriginalAirDate CallSign Name Time"
 set attributes [split $info(attributes)]
 foreach item $attributes { set info($item) "" }
 set xmldatalist [split $xmldata "\n"]
  foreach line $xmldatalist {
	foreach item $attributes {
	    if { [regexp "<$item>" $line]  } {
		set il [expr [string length $item] + 2]	
	        set sindex  [expr [string first "<$item>" $line] + $il]
	        set sindex2 [expr [string first "</$item>" $line] - 1]
	        set info($item) [string range $line $sindex $sindex2]
	    }
	}
  }
 outc "listing info parsed from xml "
  foreach item $attributes {
	outc "$item is \"$info($item)\" "
  }
}


proc setrecinfo { recfsid } {
global info attributes db
outc "entering setrecinfo"
 set progitems "Title Description EpisodeTitle OriginalAirDate"
 set progitems [split $progitems]
 set showingitems "Date Duration Time"
 set showingitems [split $showingitems]
 set recitems "StartDate StartTime StopDate StopTime"
 set recitems [split $recitems]
 set stationitems "CallSign Name"
 set stationitems [split $stationitems]

 set db [dbopen]
  set delay 0
  RetryTransaction {
	after $delay
	set delay 100
	set rec [db $db openid $recfsid]
	set showing [dbobj $rec get Showing]
	set program [dbobj $showing get Program]
	set station [dbobj $showing get Station]
	foreach item $progitems {
		if { $info($item) != "" } { dbobj $program set $item $info($item) }
	}  
	foreach item $showingitems {
		if { $info($item) != "" } { dbobj $showing set $item $info($item) }
	}
	foreach item $recitems {
		if { $info($item) != "" } { dbobj $rec set $item $info($item) }
	}
	foreach item $stationitems {
		if { $info($item) != "" } { dbobj $station set $item $info($item) }
	}
 }
}                          


proc mkblankrec { } {
outc "entering mkblankrec"
global info version
 set date  [expr [clock seconds] / 86400]
 set time  [expr [expr [clock seconds] % 86400] ]
 set db [dbopen]
 set outc stdout
 set delay 0
 RetryTransaction {
    after $delay
    set delay 100
    outc "create recording"
    set recording [db $db create Recording]
    outc "set recordingfsid"
    set recordingfsid [ dbobj $recording fsid ]
    outc "setting default info for bitrate & such"
    dbobj $recording set BitRate 0
    dbobj $recording set ErrorString "mfs_ftp insert"
    dbobj $recording set ExpirationDate [expr $date + 1]
    dbobj $recording set ExpirationTime 0
    dbobj $recording set RecordQuality 100
    dbobj $recording set Score 0
    dbobj $recording set SelectionType 3
    dbobj $recording set State 4 ;# nowshowing
    dbobj $recording set StartDate $date
    dbobj $recording set StopDate $date
    dbobj $recording set StartTime $time
    dbobj $recording set StopTime [expr $time + 1]
    outc "setting station info & misc"
    set station [db $db create Station]
    dbobj $station set CallSign "mfs_ftp"
    dbobj $station set Name     "mfs_ftp"
   
    if { $version == 3 } {
	outc "setting version 3 attributes"
        dbobj $recording set StreamFileSize 1024
        dbobj $recording set UsedBy 1
	set recordingbehavior [db $db createsub RecordingBehavior $recording]
        dbobj $recordingbehavior set DiskBehavior 1
        dbobj $recordingbehavior set PresentationBehavior 1
        dbobj $recordingbehavior set ProgramGuideBehavior 1
        dbobj $recordingbehavior set TunerBehavior 3
        dbobj $recording set RecordingBehavior $recordingbehavior
    }

 outc "create showing & set default info"
    set showing [db $db createsub Showing $recording]
    dbobj $showing set Date $date
    dbobj $showing set Duration 1
    dbobj $showing set Station $station
    dbobj $showing set Time $time
    dbobj $recording set Showing $showing
    dbobj $recording set StreamFileSize 0
 outc "create program & set default info"
    set program [db $db create Program]
    dbobj $program set Title "new recording placeholder"
    dbobj $program set OriginalAirDate "0"
    dbobj $showing set Program $program
 }
 dbclose $db
 outc "created blank recording $recordingfsid"
 return $recordingfsid
}


proc AddPart { recordingfsid streamsize } {
outc "entering AddPart"
global info
 set db [dbopen]
 set delay 0
  RetryTransaction {
    after $delay
    set delay 100
    outc "open $recordingfsid"
    set rec [db $db openid $recordingfsid]
    outc "create part"
    set part [db $db createsub RecordingPart $rec]
    outc "add part"
    dbobj $rec add Part $part
    outc "allocate [expr $streamsize / 1048576] meg"
    set fsidtwo [mfs streamcreate 131072 [expr $streamsize / 131072]]
    outc "set file to $fsidtwo"
    dbobj $part set File $fsidtwo
    dbobj $rec set StreamFileSize $info(totalsize)
  }
 dbclose $db
 return $fsidtwo
}


proc	SYST		{ args fsock } {
 set option1 [lindex $args 1] 
 tee $fsock "215 UNIX"
}


proc	TYPE		{ args fsock } {
 set option1 [lindex $args 0] 
 set option2 [lindex $args 1]
 set info(type) $option2
 tee $fsock "200 Type set to I."
}


proc	USER 	{ args fsock } {
 tee $fsock "331 User name okay, need password."
}


proc dropslash { strng } {
 if {[string index $strng 0] == "/"} {set strng [string range $strng 1 end] }
 return $strng
}


proc min {a b} {
	if {$a < $b} {return $a} else {return $b}
}


proc strim {str} {
    return [string trim $str "\{\}"]
}


proc versioncheck { } {
global version info
  outc "mfs_ftp version $info(version) - checking tivo system sw "
  set db [dbopen]
  RetryTransaction {
	set version [dbobj [db $db open "/SwSystem/ACTIVE"] get Name]
  }
  outc "version $version "
  set version [string range $version 0 1]
  if { $version == "2" } { 
	set info(mfspath) "/Recording/NowShowing" 
  } else { 
	set info(mfspath) "/Recording/NowShowingByClassic"
  }
  outc "setting mfspath to $info(mfspath) "
}


proc setup_control_socket {port} {
global control_socket info
 set control_socket(main) [socket -server newcsocket $port]
 outc "port $info(port) control socket initialized - ready for connections"
}


proc newcsocket {sock addr port} {
global control_socket info
 outc "Accept $sock from $addr port $port "
 abortcheck
 set control_socket(addr,$sock) [list $addr $port]
 if { $info(multi_c_threads) == 0 } {
	close $control_socket(main)
 }
 fconfigure $sock -buffering line
 tee $sock "220 mfs_ftp.tcl server $info(version) - connect from $addr $port"
 fileevent $sock readable [list readlinefromsocket $sock]
 set info(dir) "/"
 set sinfo [fconfigure $sock -sockname]
 # outc "server ip is \"$sinfo\""
 set sinfo [split $sinfo \ ]
 set sinfo [lindex $sinfo 0]
 regsub -all \\. $sinfo , info(localip)
 # outc "server ip is \"$info(localip)\" "
}


proc readlinefromsocket {sock} {
global control_socket info
 if {[eof $sock] || [catch {gets $sock line}]} {
	outc "control connection lost - closing socket"
	catch { close $sock }
	outc "$control_socket(addr,$sock) closed"
	catch { unset control_socket(addr,$sock) }
 	if { $info(multi_c_threads) == 0 } {
		outc "setting up new control socket"
 		after 1 setup_control_socket $info(port)
}
 } else {
	outc "echo to verify: $line"
	parseline $line $sock
 }
}


proc parseline { line fsock } {
global control_socket 
 set args [split $line]
 set cmd [lindex $args 0] 
 set info(lastsock) $fsock
 abortcheck
  switch $cmd {
	ABOR		{ ABOR	$args $fsock }
	USER 		{ USER	$args $fsock }
	PASS		{ PASS	$args $fsock }
	CWD		{ CWD	$args $fsock }
	CDUP		{ CDUP	$args $fsock }
	DELE		{ DELE  $args $fsock $line }
	PORT		{ PORT	$line $fsock }
	PASV		{ PASV	$args $fsock }
        QUIT            { QUIT $fsock }
	TYPE		{ TYPE	$args $fsock }
	MODE		{ MODE	$args $fsock }
	RETR		{ RETR	$args $fsock $line }
	STOR		{ STOR	$args $fsock $line }
	PWD 		{ PWD 	$args $fsock }
	LIST		{ LIST	$line $fsock }
	SYST		{ SYST	$args $fsock }
	NOOP		{ NOOP	$args $fsock }
        NLST            { LIST  $args $fsock }
	exit		{ exitproc $fsock raw_ftp_command_exit }
	EXIT		{ exitproc $fsock raw_ftp_command_EXIT }
	PHEONIX		{ pheonix_mfs_ftp $fsock raw_ftp_command_PHEONIX }
	default 	{ default $args $fsock }
  }
}


proc default { args fsock } {
	set option1 "[lindex $args 0]"
        if { [regexp {ABOR} $option1]  } {
		set args [lreplace $args 0 1 ABOR]
		ABOR $args $fsock
		return
	} else {
		tee $fsock "502 Command not implemented"
	}
}


proc isInt { val } {
    return [regexp {^-?[1-9][0-9]*$|^-?0[xX][0-9a-fA-F]+$|^-?0[0-7]*$} $val]
}
  

proc abortcheck { } {
  if [file exist ./abort.txt] {
	puts stdout "found abort.txt, exiting"
	exit
  }
}


proc exitproc { sock callingproc } {
global control_socket info
	tee  $sock "200 shutting down the server NOW from \"$callingproc\""
	catch { close $control_socket(main)}
	catch {close $sock}
	set forever end
	exit
}


proc pheonix_mfs_ftp { fsock callingproc } {
global info
 tee $fsock "$callingproc quiting mfs_ftp.tcl and forking a new one"
 after 200
 exec /var/mfs_ftp/mfs_ftp.tcl $info(port) $info(wrapped) 2>& stdout &
 exitproc stdout pheonix_mfs_ftp
}      


proc fork_mfs_ftp { } {
global info
 outc "\n forking a new copy of mfs_ftp "
 exec /var/mfs_ftp/mfs_ftp.tcl $info(port) $info(wrapped) 2>& stdout &
}


proc ping {} {
global info
 outc "ping - mfs_ftp running on $info(localip) : $info(port) "
 abortcheck
 after $info(pingpong) pong
}


proc pong {} {
global info
 outc "pong - mfs_ftp running on $info(localip) : $info(port) "
 abortcheck
 after $info(pingpong) ping
}


proc bgerror { error } {
global info
 puts stdout "bgerror invoked with error \n \n \" $error \" \n"
 puts stdout "re-initializing mfs_ftp\n"
 puts stdout "close the currnet ftp connection and simply open another\n"
 if { $info(wrapped) == "wrapped" } { exitproc stdout bgerror }
 after 500 reset_mfs_ftp
}


proc get_tzoffset {mfstz dst} {
	if { $mfstz <= 0 } {
		set tz $mfstz
	} else {
		set tzlist "-5 -6 -7 -8 -9 -10 0 1 2 3 4 5 6 7 8 9 10 11 12 -1 -2 -3 -4 -11 -12"
                set tz [lindex $tzlist [expr $mfstz - 1]]
	}
	if { $dst == 2 } {
		set date [clock format [clock seconds] -format "%1d %w %1m %1H %1M"]
		scan $date "%d %d %d %d %d" dom dow month hour min
		set dlsval 0
		if {$::uktivo} {
			if {$month > 3 && $month < 10} {
				set dlsval 1
			} elseif {$month == 3 && $dom >= 25 && $dow == 0 && $hour >= 1} {

			set dlsval 1
		} elseif {$month == 3 && $dom >= 25 && $dow != 0 && ($dom-24-$dow >= 1) } {
			set dlsval 1
		}
	} else {
		if {$month > 4 && $month < 10} {
			set dlsval 1
		} elseif {$month == 4 && $dom > 7} {
			set dlsval 1
		} elseif {$month == 4 && $dom <= 7 && $dow == 0 && $hour >= 2} {
			set dlsval 1
		} elseif {$month == 4 && $dom <= 7 && $dow != 0 && ($dom-$dow > 0)} {
			set dlsval 1
		}
	}
	if {$dlsval == 0} {
		if {$month == 10 && $dom < 25} {
			set dlsval 1
		} elseif {$month == 10 && $dom >= 25 && $dow == 0 && $hour < 2} {
			set dlsval 1
		} elseif {$month == 10 && $dom >= 25 && $dow != 0 && ($dom-24-$dow < 1) } {
			set dlsval 1
		}
	}
	if {$dlsval == 1} {
		return [expr ($tz+1)*60*60]
	} else {
		return [expr $tz*60*60]
	}
	} else {
		return [expr $tz*60*60]
	}
}


proc get_tzoffset_one { } {
global info tzoffset tivoswversion version3 uktivo
set db [dbopen]
set delay 0
RetryTransaction {
	set swsystem [db $db open /SwSystem/ACTIVE]
	set tivoswversion [dbobj $swsystem get Name]
	set setup [db $db open /Setup]
	if { [string range $tivoswversion 0 2] >= 3.0 } {
		set version3 1
	} else {
		set version3 0
	}
	if {[PrefixMatches "2.5.5" $::tivoswversion]} {
		set uktivo 1
	} else {
		set uktivo 0
	}
	set suffix [string range $tivoswversion [expr [string length $tivoswversion] - 3] end]
	if { [lsearch "001 011 031" $suffix] >= 0 } {
		set dtivo 1
	} else {
		set dtivo 0
	}
	if { [PrefixMatches "2.5-" $tivoswversion] } {
		#puts stdout "Error: This version is not supported"
		#exit
		dbclose $db
		set datestr "Dec 25 2001"
	}
	if {$::version3} {
		set lconfig  [db $db open /State/LocationConfig]
		set setup [db $db open /State/ServiceConfig]
		set setuptz [dbobj $lconfig get TimeZoneOld]
		set daylightsavings [dbobj $lconfig get DaylightSavingsPolicy]
	} else {
		set setuptz [dbobj $setup get TimeZone]
		set daylightsavings [dbobj $setup get DaylightSavingsPolicy]
	}
	if {$setuptz == ""} {
		set setuptz 0
	}
	if {$daylightsavings == ""} {
		set daylightsavings 2
	}
}
set tzoffset [get_tzoffset $setuptz $daylightsavings]
dbclose $db
return $tzoffset
}


init_mfs_ftp
set forever start
vwait forever  