/* a module to convert a US NTSC based TiVo to PAL input and output

   tridge@samba.org, 2000
   
   This version is for a US based Standalone Tivo (Philips HDR xxx) 
   running V3.0 SW
   modified by STS for 3.0 SW
   many clues and changes based on the work done by d18c7db
*/

#define DEBUG	0	/* turn on debug prints/statements */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/fs.h>

#include "tv_chan.h"  // get tuner channel/frequency info for respective area

#define TUNER_BUSADDR		0xC2
#define VERSION				"3.0"

/* offsets into modules
These offsets have to be reajusted for every new version of the Tivo SW
It took me a while to find out, what Tridge did, so here is a recipe: 
run readelf -r <module>.o 
search for the function name column "Symbol's Name" 
unfortunately V3.0 SW does not export most of these names any more,
so this is a comparisson of addresses and guess,
which .text could be the right one.
I use the syntax:
_call_		if this function is called in the module
_export_	if this function is exproted by the module
*/

/* offsets into fpga7114.o */
#define offset_call_fpga_hw_ioctl		0x298c
#define offset_call_MSPInit				0x64B8
#define offset_call_fpga_get_channel	0x3b44
#define offset_export_fpga_hw_ioctl		0x4C

#define offset_standin_tune		0x0000	/* not fixed yet, can't find them in 3.0 */
#define offset_BangRegs			0x0000	/* not fixed yet, can't find them in 3.0 */
#define offset_getFreq_C2		0x0000	/* not fixed yet, can't find them in 3.0 */
#define offset_getFreq_C6		0x0000	/* not fixed yet, can't find them in 3.0 */

/* reference to functions from the I2C module */
extern int I2CRead(int module, int subaddress, unsigned char *data, int len);
extern int I2CReadNoSub(int module, unsigned char *data, int len);
extern int I2CWrite(int module, int subaddress, unsigned char *data, int len);
extern int I2CWriteNoSub(int module, unsigned char *data, int len);
/* end of external references */

static u32 bypass_offset(u32 addr);


/* useful to know what channel we are on */
static int channel = -1;

static int iicw4(int module, int subaddress, u32 v)
{
	return I2CWrite(module, subaddress, (char *)&v, 4);
}

static int iicw1(int module, int subaddress, unsigned char v)
{
	return I2CWrite(module, subaddress, (char *)&v, 1);
}

/* not needed any more
static int iicw1d(int mod, int sub, unsigned char v)
{
	unsigned char old;

	I2CRead(mod, sub, &old, 1);
	if (v == old)
		return 0;
	return I2CWrite(mod, sub, &v, 1);
}
*/


/* not needed any more: read one byte to a i2c module
static unsigned char iicr1(int module, int subaddress)
{
	unsigned char v;
	I2CRead(module, subaddress, &v, 1);
	return v;
}
*/

static void cache_flush(u32 addr)
{
	__asm ("dcbf 0,%0\n\t"
	       "sync \n\t"
	       "icbi 0,%0 \n\t"
	       "isync \n\t"
	       : : "r" (addr));
}

static int channel;

/* we should be using frequencyAdjustmentG but it doesn't seem to work!? */
static int adjust;


/*
 * a hook for the Tmdh2Tune fn.
 * We need to tune to channel `chan' with fine-tune offset
 * frequencyAdjustmentG.
 */
static int Tmdh2Tune_x(int chan, int sigtype)
{
	int i, f;
	static int last_channel = -1;

#if	DEBUG
	printk("Tmdh2Tune_x channel=%d\n", channel);
#endif
	if (last_channel != chan) {
		adjust = 0;
	}

	for (i = 0; tv_chans[i].freq != 0; i++)
		if (chan == tv_chans[i].chan)
			break;
	if (tv_chans[i].freq == 0)
		return -1;
	f = (tv_chans[i].freq + 38900) + adjust * 125 / 4;
	f = ((f * 16 + 500) / 1000) << 16 | 0xce00;

//	if (chan != last_channel)
//		f = ((f * 16 + 500) / 1000) << 16 | 0xce00;
//	else
//		f = ((f * 32 + 500) / 1000) << 16 | 0x8a00;

	if (tv_chans[i].freq < 170000)
		f |= 0xa0;
	else if (tv_chans[i].freq < 450000)
		f |= 0x90;
	else
		f |= 0x30;
	I2CWriteNoSub(TUNER_BUSADDR, (u8 *)&f, 4);
	iicw1(0x42, 0x2, 0x80);
#if DEBUG
	printk("tuned to 0x%08x for chan %d adj=%d\n", 
	       (unsigned)f, chan, adjust);
#endif
	last_channel = chan;
	return 0;
}

static int fpga_hw_ioctl_x(u32 a1, u32 a2, u32 a3, u32 a4)
{
	int (*fpga_hw_ioctl)(u32 a1, u32 a2, u32 a3, u32 a4) = bypass_offset(offset_call_fpga_hw_ioctl);
	int ret;
#if DEBUG
	printk("fpga_hw_ioctl: a1=%x a2=%x a3=%x a4=%x\n", 
	       a1, a2, a3, a4);
#endif
	if (a3 == 0x8000) {
		u32 *p = (u32 *)a4;
		channel = p[1];
#if DEBUG
		printk("8000: %x %x %x %x\n", p[0], p[1], p[2], p[3]);
		printk("channel=%d\n", channel);
#endif	
		p[1] = ~0;
		p[3] = 0x220;
	}
	ret = fpga_hw_ioctl(a1, a2, a3, a4);

	if (a3 == 0x8000) {
		Tmdh2Tune_x(channel, 0);
	}
	return ret;
}




/* a hook for the MSPInit function */
static void MSPInit_x(int x)
{
	/* this gets the vertical sizing write on the screen. Not
	 * really part of the msp init, but this is an easy place to
	 * put it */
	iicw1(0x8c, 0x7a, 0x5);
	iicw1(0x8c, 0x7b, 0x38);

#if DEBUG
	printk("MSP channel=%d\n", channel);
#endif
	if ((channel == -1) || (channel == -2)) {
		void (*MSPInit)(int) = bypass_offset(offset_call_MSPInit);
		MSPInit(0);
		/* the UK tivo is wired differently from the US tivo -
		   this directs the correct SCART input to be used for compositte
		   audio input when we are not using the tuner
		   channel = -1 means composite Video, Channel = -2 means S-Video */
		iicw4(0x88, 0x12, 0x00130300);
		return;
	}
	iicw4(0x88, 0x10, 0x00300060);
	iicw4(0x88, 0x12, 0x000b0220);
	iicw4(0x88, 0x12, 0x00130000);
	iicw4(0x88, 0x12, 0x000d1900);
}

static u32 bypass_offset(u32 addr)
{
	extern int fpga_get_channel(void);
	
#if DEBUG
	printk("&fpga_get_channel= %d\n", channel);
#endif
	
	return addr + (((u32)fpga_get_channel)- offset_call_fpga_get_channel);
}

static void install_bypass_fn(u32 newfn, u32 call)
{
	u32 old;
	call = bypass_offset(call);
	old = *(u32 *)call;
	*(u32 *)call = (0x48000001) | ((newfn - call)&0xFFFFFF);
	cache_flush(call);
}

int init_module(void)
{
	install_bypass_fn((u32)fpga_hw_ioctl_x, offset_export_fpga_hw_ioctl);

	{
		/* run readelf -r fpga7114.o 
		search for ALL instances of MSPInit in the column "Symbol's Name" 
		The Offset in the first column is what you are looking for 
		unfortunately V3.0 SW does not export these names any more,
		so this is a comparisson of addresses and guess,
		which .text could be the right one */

		u32 addrs[] = {0x1be8, 0x1c00, 0x1c18, 0x1c34, 0x222c,
			       0x2244, 0x225c, 0x2288, 0x712c, 0};
		int i;
		for (i = 0; addrs[i]; i++) {
			install_bypass_fn((u32)MSPInit_x, addrs[i]);
		}
	}


	return 0;
}

void cleanup_module(void)
{
	printk("palmod removed!\n");
}
