/* a module to convert a US NTSC based TiVo to PAL input and output

   tridge@samba.org, 2000
*/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/fs.h>

extern int I2CWrite(int module, int subaddress, unsigned char *data, int len);

static int iicw4(int module, int subaddress, u32 v)
{
	return I2CWrite(module, subaddress, (char *)&v, 4);
}

static int getFreq_x(int chan, int u1, u32 *p1, u32 *p2)
{
	u32 f;
	int i;
	u32 freq;

	struct {
		int chan;
		u32 freq;
	} chans[] = {
    { 0,	 46250 },
    { 1,	 57250 },
    { 2,	 64250 },
    { 3,	 86250 },
    { 4,  	 95250 },
    { 5,  	102250 },
    { 6,  	175250 },
    { 7,  	182250 },
    { 8,  	189250 },
    { 9,  	196250 },
    { 10, 	209250 },
    { 11,	216250 },
    { 28,	527250 },
    { 29,	534250 },
    { 30,	541250 },
    { 31,	548250 },
    { 32,	555250 },
    { 33,	562250 },
    { 34,	569250 },
    { 35,	576250 },
    { 36,       591250 },
    { 39,	604250 },
    { 40,	611250 },
    { 41,	618250 },
    { 42,	625250 },
    { 43,	632250 },
    { 44,	639250 },
    { 45,	646250 },
    { 46,	653250 },
    { 47,	660250 },
    { 48,	667250 },
    { 49,	674250 },
    { 50,	681250 },
    { 51,	688250 },
    { 52,	695250 },
    { 53,	702250 },
    { 54,	709250 },
    { 55,	716250 },
    { 56,	723250 },
    { 57,	730250 },
    { 58,	737250 },
    { 59,	744250 },
    { 60,	751250 },
    { 61,	758250 },
    { 62,	765250 },
    { 63,	772250 },
    { 64,	779250 },
    { 65,	786250 },
    { 66,	793250 },
    { 67,	800250 },
    { 68,	807250 },
    { 69,	814250 },
    {0,0}
	};

//	printk("getFreq_x chan=%d u1=%d\n", chan, u1);

	for (i=0;chans[i].freq;i++) {
		if (chans[i].chan == chan) break;
	}

	freq = chans[i].freq;

	if (freq == 0) {
		f = 0;
	} else {
		f = (freq-471250)*1000 + 510150000;
	}
	
	*p1 = f;
	if (chans[i].freq > 385250) {
		*p2 = 2;
	} else if (chans[i].freq > 137250) {
		*p2 = 1;
	} else {
		*p2 = 0;
	}

	/* setup Australian sound - see page 22 of msp34x0g_3pd.pdf */
	iicw4(0x88, 0x10, 0x00200003); /* B/G Dual FM Stereo */

	return 0;
}

static void cache_flush(u32 addr)
{
	__asm ("dcbf 0,%0\n\t"
	       "sync \n\t"
	       "icbi 0,%0 \n\t"
	       "isync \n\t"
	       : : "r" (addr));
}

static void install_bypass(u32 base, u32 newfn, u32 call)
{
	u32 old;
	call += base;
	old = *(u32 *)call;
	*(u32 *)call = (0x48000001) | ((newfn - call)&0xFFFFFF);
	cache_flush(call);
	printk("palmod: patched at 0x%x with 0x%x (was 0x%x)\n", 
	       call, *(u32 *)call, old);
}

static void install_bypass_fn(u32 newfn, u32 call)
{
	extern int fpga_get_channel(void);
	install_bypass(((u32)fpga_get_channel)-0x3a60, newfn, call);
}

int init_module(void)
{
	install_bypass_fn((u32)getFreq_x, 0x73b4);

	printk("ukchan installed\n");

	return 0;
}

void cleanup_module(void)
{
	printk("ukchan removed!\n");
}

