/* a module to convert a US NTSC based TiVo to PAL input and output

   tridge@samba.org, 2000
*/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/fs.h>

#ifdef PAULUS
#define TUNER_BUSADDR	0x84
#else
#define TUNER_BUSADDR	0xc2
#endif /* PAULUS */

extern int I2CRead(int module, int subaddress, unsigned char *data, int len);
extern int I2CReadNoSub(int module, unsigned char *data, int len);
extern int I2CWrite(int module, int subaddress, unsigned char *data, int len);
extern int I2CWriteNoSub(int module, unsigned char *data, int len);
static u32 bypass_offset(u32 addr);

static int iicw4(int module, int subaddress, u32 v)
{
	return I2CWrite(module, subaddress, (char *)&v, 4);
}

static int iicw1(int module, int subaddress, unsigned char v)
{
	return I2CWrite(module, subaddress, (char *)&v, 1);
}

static int iicw1d(int mod, int sub, unsigned char v)
{
	unsigned char old;

	I2CRead(mod, sub, &old, 1);
	if (v == old)
		return 0;
	return I2CWrite(mod, sub, &v, 1);
}

/* read one byte to a i2c module */
static unsigned char iicr1(int module, int subaddress)
{
	unsigned char v;
	I2CRead(module, subaddress, &v, 1);
	return v;
}

static struct {
	int chan;
	u32 freq;
} oz_chans[] = {
    { 0,	 46250 },
    { 1,	 57250 },
    { 2,	 64250 },
    { 3,	 86250 },
    { 4,  	 95250 },
    { 5,  	102250 },
    { 6,  	175250 },
    { 7,  	182250 },
    { 8,  	189250 },
    { 9,  	196250 },
    { 10, 	209250 },
    { 11,	216250 },
    { 28,	527250 },
    { 29,	534250 },
    { 30,	541250 },
    { 31,	548250 },
    { 32,	555250 },
    { 33,	562250 },
    { 34,	569250 },
    { 35,	576250 },
    { 36,       591250 },
    { 39,	604250 },
    { 40,	611250 },
    { 41,	618250 },
    { 42,	625250 },
    { 43,	632250 },
    { 44,	639250 },
    { 45,	646250 },
    { 46,	653250 },
    { 47,	660250 },
    { 48,	667250 },
    { 49,	674250 },
    { 50,	681250 },
    { 51,	688250 },
    { 52,	695250 },
    { 53,	702250 },
    { 54,	709250 },
    { 55,	716250 },
    { 56,	723250 },
    { 57,	730250 },
    { 58,	737250 },
    { 59,	744250 },
    { 60,	751250 },
    { 61,	758250 },
    { 62,	765250 },
    { 63,	772250 },
    { 64,	779250 },
    { 65,	786250 },
    { 66,	793250 },
    { 67,	800250 },
    { 68,	807250 },
    { 69,	814250 },
    {0,0}
};

static void cache_flush(u32 addr)
{
	__asm ("dcbf 0,%0\n\t"
	       "sync \n\t"
	       "icbi 0,%0 \n\t"
	       "isync \n\t"
	       : : "r" (addr));
}

#ifdef PAULUS
static void paulus_write_tuner(int freq)
{
	int x, i;

	x = iicr1(0x42, 0x12);
	x = (x & 0xf) | 0x10;
	iicw1(0x42, 0x12, x);

	iicw1(0x84, 1, 0x10);
	iicw1(0x42, 0x11, 0x40); /* frame high */
	
	for (i = 18; i >= 0; --i) {
		int y = ((freq >> i) & 1) << 2;
		iicw1(0x84, 1, y + 0x10); /* set data value */
		iicw1(0x84, 1, y + 0x11); /* clock high */
		iicw1(0x84, 1, y + 0x10); /* clock low */
	}
	iicw1(0x84, 1, 0x10);
	iicw1(0x42, 0x11, 0);	/* frame low */
}
#endif /* PAULUS */

/* we should be using frequencyAdjustmentG but it doesn't seem to work!? */
static int adjust;

/*
 * a hook for the Tmdh2Tune fn.
 * We need to tune to channel `chan' with fine-tune offset
 * frequencyAdjustmentG.
 */
static int Tmdh2Tune_x(int chan, int sigtype)
{
	int i, f;
	static int last_channel = -1;
#ifdef PAULUS
	static int last_stereo = 0;
#endif
	
	if (last_channel != chan) {
		adjust = 0;
	}

	for (i = 0; oz_chans[i].freq != 0; i++)
		if (chan == oz_chans[i].chan)
			break;
	if (oz_chans[i].freq == 0)
		return -1;
	f = (oz_chans[i].freq + 38900) + adjust * 125 / 4;
#ifdef PAULUS
	f /= 50;
	if (f >= 8212)
		f |= 0x40000;
	else /*if (f >= 3552)*/
		f |= 0x10000;

	paulus_write_tuner(f);

	if (chan != last_channel) {
		iicw1(0x84, 0, 0x10);	/* assume mono for now */
		last_stereo = 0;
	} else {
		/* test for stereo sound */
		i = iicr1(0x84, 0) & 0x40;
		if (i != last_stereo) {
			iicw1(0x84, 0, (i? 0x2a: 0x10));
			last_stereo = i;
		}
	}
#else
	f = ((f * 16 + 500) / 1000) << 16 | 0xce00;
//	if (chan != last_channel)
//		f = ((f * 16 + 500) / 1000) << 16 | 0xce00;
//	else
//		f = ((f * 32 + 500) / 1000) << 16 | 0x8a00;
	if (oz_chans[i].freq < 170000)
		f |= 0xa0;
	else if (oz_chans[i].freq < 450000)
		f |= 0x90;
	else
		f |= 0x30;
	I2CWriteNoSub(TUNER_BUSADDR, (u8 *)&f, 4);
#if 0
	printk("tuned to 0x%08x for chan %d adj=%d\n", 
	       (unsigned)f, chan, adjust);
#endif
#endif
	last_channel = chan;
	return 0;
}

/* a hook for the MSPInit function */
static void MSPInit_x(int x)
{
	/* this gets the vertical sizing write on the screen. Not
	 * really part of the msp init, but this is an easy place to
	 * put it */
	iicw1(0x8c, 0x7a, 0x5);
	iicw1(0x8c, 0x7b, 0x38);

	if (!x) {
		void (*MSPInit)(int) = bypass_offset(0x6370);
		MSPInit(0);
		/* the UK tivo is wired differently from the US tivo -
		   this directs the correct SCART input to be used for compositte
		   audio input when we are not using the tuner */
		iicw4(0x88, 0x12, 0x00130300);
		return;
	}
	iicw4(0x88, 0x10, 0x00300060);
	iicw4(0x88, 0x12, 0x000b0220);
	iicw4(0x88, 0x12, 0x00130000);
	iicw4(0x88, 0x12, 0x000d1900);
}

static u32 bypass_offset(u32 addr)
{
	extern int fpga_get_channel(void);
	return addr + (((u32)fpga_get_channel)-0x3a60);
}

static void install_bypass_fn(u32 newfn, u32 call)
{
	u32 old;
	call = bypass_offset(call);
	old = *(u32 *)call;
	*(u32 *)call = (0x48000001) | ((newfn - call)&0xFFFFFF);
	cache_flush(call);
}

#ifndef PAULUS
static void TmdhGetTune_x(int *toolow, int *nottoohigh)
{
	unsigned char status = 0;

	I2CReadNoSub(TUNER_BUSADDR, &status, 1);
	*toolow = (status & 0x7) > 2;
	*nottoohigh = (status & 7) >= 2;
	if (*toolow) {
		adjust += 1;
		if (adjust > 50) adjust = -50;
	} else if (! *nottoohigh) {
		adjust -= 1;
		if (adjust < -50) adjust = 50;
	}
}
#endif

int init_module(void)
{
	install_bypass_fn((u32)Tmdh2Tune_x, 0x7550);
#ifndef PAULUS
	install_bypass_fn((u32)TmdhGetTune_x, 0x2b64);
	install_bypass_fn((u32)TmdhGetTune_x, 0x82cc);
	install_bypass_fn((u32)TmdhGetTune_x, 0x837c);
	install_bypass_fn((u32)TmdhGetTune_x, 0x8400);
#endif

	{
		u32 addrs[] = {0x1b08, 0x1b20, 0x1b38, 0x1b54, 0x2148,
			       0x2160, 0x2178, 0x21a4, 0x7028, 0};
		int i;
		for (i = 0; addrs[i]; i++) {
			install_bypass_fn((u32)MSPInit_x, addrs[i]);
		}
	}


	return 0;
}

void cleanup_module(void)
{
	printk("palmod removed!\n");
}
